/*
 * libhpcables: hand-helds support libraries.
 * Copyright (C) 2013 Lionel Debroux
 * Code patterns and snippets borrowed from libticables & libticalcs:
 * Copyright (C) 1999-2009 Romain Livin
 * Copyright (C) 2009-2013 Lionel Debroux
 * Copyright (C) 1999-2013 libti* contributors.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * \file hpcables.c Cables: base part.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <hpcables.h>
#include "logging.h"
#include "error.h"

#include <stdlib.h>
#include <hidapi/hidapi.h>

extern const cable_fncts cable_prime_hid_fncts;

const cable_fncts * hpcables_all_cables[CABLE_MAX] = {
    NULL,
    &cable_prime_hid_fncts
};


// not static, must be shared between instances
int hpcables_instance_count = 0;

HPEXPORT int HPCALL hpcables_init(void (*log_callback)(const char *format, va_list args)) {
    int res;

    hpcables_log_set_callback(log_callback);
    hpcables_info("hpcables library version %s compiled on " __DATE__ " " __TIME__, hpcables_version_get());

    res = hid_init();
    if (res == 0) {
        res = ERR_SUCCESS;
        hpcables_info("%s: init succeeded", __FUNCTION__);
    }
    else {
        res = ERR_LIBRARY_INIT;
        hpcables_error("%s: init failed", __FUNCTION__);
    }
    return res;
}

HPEXPORT int HPCALL hpcables_exit(void) {
    int res;
    hid_exit();
    res = ERR_SUCCESS;
    hpcables_info("%s: exit succeeded", __FUNCTION__);
    return res;
}


HPEXPORT const char* HPCALL hpcables_version_get (void) {
    return VERSION;
}


HPEXPORT cable_handle * HPCALL hpcables_handle_new(cable_model model) {
    cable_handle * handle = NULL;
    if (model > CABLE_NONE && model < CABLE_MAX) {
        handle = (cable_handle *)calloc(1, sizeof(*handle));

        if (handle != NULL) {
            handle->model = model;
            handle->fncts = hpcables_all_cables[model];
            hpcables_info("%s: handle allocation succeeded", __FUNCTION__);
        }
        else {
            hpcables_error("%s: handle allocation failed", __FUNCTION__);
        }
    }
    else {
        hpcables_error("%s: invalid model", __FUNCTION__);
    }

    return handle;
}

HPEXPORT int HPCALL hpcables_handle_del(cable_handle * handle) {
    int res;
    if (handle != NULL) {
        free(handle->handle);
        handle->handle = NULL;

        free(handle);
        handle = NULL;
        res = ERR_SUCCESS;
        hpcables_info("%s: handle deletion succeeded", __FUNCTION__);
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }

    return res;
}

HPEXPORT int HPCALL hpcables_handle_display(cable_handle * handle) {
    int res;
    if (handle != NULL) {
        hpcables_info("Link cable handle details:");
        hpcables_info("\tmodel: %s", hpcables_model_to_string(handle->model));
        hpcables_info("\tread_timeout: %d", handle->read_timeout);
        hpcables_info("\topen: %d", handle->open);
        hpcables_info("\tbusy: %d", handle->busy);
        res = ERR_SUCCESS;
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }

    return res;
}

HPEXPORT cable_model HPCALL hpcables_get_model(cable_handle * handle) {
    cable_model model = CABLE_NONE;
    if (handle != NULL) {
        model = handle->model;
    }
    else {
        hpcalcs_error("%s: handle is NULL", __FUNCTION__);
    }

    return model;
}

#define DO_BASIC_HANDLE_CHECKS() \
    if (!handle->open) { \
        res = ERR_CABLE_NOT_OPEN; \
        hpcalcs_error("%s: cable not open", __FUNCTION__); \
        break; \
    } \
    DO_BASIC_HANDLE_CHECKS2()

#define DO_BASIC_HANDLE_CHECKS2() \
    if (handle->busy) { \
        res = ERR_CABLE_BUSY; \
        hpcalcs_error("%s: cable busy", __FUNCTION__); \
        break; \
    } \
    if (handle->fncts == NULL) { \
        res = ERR_CABLE_INVALID_FNCTS; \
        hpcalcs_error("%s: fncts is NULL", __FUNCTION__); \
        break; \
    }

HPEXPORT int HPCALL hpcables_options_get_read_timeout(cable_handle * handle) {
    int timeout = 0;
    if (handle != NULL) {
        timeout = handle->read_timeout;
    }
    else {
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }
    return timeout;
}

HPEXPORT int HPCALL hpcables_options_set_read_timeout(cable_handle * handle, int read_timeout) {
    int res;
    if (handle != NULL) {
        do {
            int (*set_read_timeout) (cable_handle *, int);

            DO_BASIC_HANDLE_CHECKS2()

            set_read_timeout = handle->fncts->set_read_timeout;
            if (set_read_timeout != NULL) {
                handle->busy = 1;
                res = (*set_read_timeout)(handle, read_timeout);
                if (res == ERR_SUCCESS) {
                    hpcables_info("%s: set_read_timeout succeeded", __FUNCTION__);
                }
                else {
                    hpcables_error("%s: set_read_timeout failed", __FUNCTION__);
                }
                handle->busy = 0;
            }
            else {
                res = ERR_CABLE_INVALID_FNCTS;
                hpcables_error("%s: fncts->set_read_timeout is NULL", __FUNCTION__);
            }
        } while (0);
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }
    return res;
}

HPEXPORT int HPCALL hpcables_cable_open(cable_handle * handle) {
    int res;
    if (handle != NULL) {
        do {
            int (*open) (cable_handle *);

            if (handle->open) {
                res = ERR_CABLE_OPEN;
                hpcables_error("%s: cable already open", __FUNCTION__);
                break;
            }
            DO_BASIC_HANDLE_CHECKS2()

            open = handle->fncts->open;
            if (open != NULL) {
                handle->busy = 1;
                res = (*open)(handle);
                if (res == ERR_SUCCESS) {
                    handle->open = 1;
                    hpcables_info("%s: open succeeded", __FUNCTION__);
                }
                else {
                    hpcables_error("%s: open failed", __FUNCTION__);
                }
                handle->busy = 0;
            }
            else {
                res = ERR_CABLE_INVALID_FNCTS;
                hpcables_error("%s: fncts->open is NULL", __FUNCTION__);
            }
        } while (0);
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }
    return res;
}

HPEXPORT int HPCALL hpcables_cable_close(cable_handle * handle) {
    int res;
    if (handle != NULL) {
        do {
            int (*close) (cable_handle *);

            DO_BASIC_HANDLE_CHECKS()

            close = handle->fncts->close;
            if (close != NULL) {
                handle->busy = 1;
                res = (*close)(handle);
                if (res == ERR_SUCCESS) {
                    handle->open = 0;
                    hpcables_info("%s: close succeeded", __FUNCTION__);
                }
                else {
                    hpcables_error("%s: close failed", __FUNCTION__);
                }
                handle->busy = 0;
            }
            else {
                res = ERR_CABLE_INVALID_FNCTS;
                hpcables_error("%s: fncts->close is NULL", __FUNCTION__);
            }
        } while (0);
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }
    return res;
}

HPEXPORT int HPCALL hpcables_cable_send (cable_handle * handle, uint8_t * data, uint32_t len) {
    int res;
    if (handle != NULL) {
        do {
            int (*send) (cable_handle *, uint8_t *, uint32_t);

            DO_BASIC_HANDLE_CHECKS()

            send = handle->fncts->send;
            if (send != NULL) {
                handle->busy = 1;
                res = (*send)(handle, data, len);
                if (res == ERR_SUCCESS) {
                    //hpcables_info("%s: send succeeded", __FUNCTION__);
                }
                else {
                    hpcables_warning("%s: send failed", __FUNCTION__);
                }
                handle->busy = 0;
            }
            else {
                res = ERR_CABLE_INVALID_FNCTS;
                hpcables_error("%s: fncts->send is NULL", __FUNCTION__);
            }
        } while (0);
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }
    return res;
}

HPEXPORT int HPCALL hpcables_cable_recv (cable_handle * handle, uint8_t * data, uint32_t * len) {
    int res;
    if (handle != NULL) {
        do {
            int (*recv) (cable_handle *, uint8_t *, uint32_t *);

            DO_BASIC_HANDLE_CHECKS()

            recv = handle->fncts->recv;
            if (recv != NULL) {
                handle->busy = 1;
                res = (*recv)(handle, data, len);
                if (res == ERR_SUCCESS) {
                    //hpcables_info("%s: recv succeeded", __FUNCTION__);
                }
                else {
                    hpcables_warning("%s: recv failed", __FUNCTION__);
                }
                handle->busy = 0;
            }
            else {
                res = ERR_CABLE_INVALID_FNCTS;
                hpcables_error("%s: fncts->recv is NULL", __FUNCTION__);
            }
        } while (0);
    }
    else {
        res = ERR_INVALID_HANDLE;
        hpcables_error("%s: handle is NULL", __FUNCTION__);
    }
    return res;
}

#undef DO_BASIC_HANDLE_CHECKS2
#undef DO_BASIC_HANDLE_CHECKS
